@@ -1,6 +1,6 @@
 /* crlfetch.c - LDAP access
  *      Copyright (C) 2002 Klarlvdalens Datakonsult AB
- *      Copyright (C) 2003 g10 Code GmbH
+ *      Copyright (C) 2003, 2004 g10 Code GmbH
  *
  * This file is part of DirMngr.
  *
@@ -60,7 +60,7 @@ fun_closer (void *cookie)
   return 0;
 }
 
-/* Helper to prepare a stream from VALUE and VALUELNE and return that
+/* Helper to prepare a stream from VALUE and VALUELEN and return that
    stream in R_FP.  In the error case VALUE will be freed. */
 static gpg_error_t
 setup_funopen (unsigned char *value, size_t valuelen, FILE **r_fp)
@@ -107,21 +107,41 @@ crl_fetch (const char *url, FILE **r_fp)
   unsigned char *value = NULL;
   size_t valuelen;
   parsed_uri_t uri;
+  char *free_this = NULL;
 
   *r_fp = NULL;
 
   err = http_parse_uri (&uri, url);
   http_release_parsed_uri (uri);
+  if (err && url && !strncmp (url, "https:", 6))
+    {
+      /* Our HTTP code does not support TLS, thus we can't use this
+         scheme and it is frankly not useful for CRL retrieval anyway.
+         We resort to using http, assuming that the server also
+         provides plain http access. */
+      free_this = xtrymalloc (strlen (url) + 1);
+      if (free_this)
+        {
+          strcpy (stpcpy (free_this,"http:"), url+6);
+          err = http_parse_uri (&uri, free_this);
+          http_release_parsed_uri (uri);
+          if (!err)
+            {
+              log_info (_("using \"http\" instead of \"https\"\n"));
+              url = free_this;
+            }
+        }
+    }
   if (!err) /* Yes, our HTTP code groks that. */
     {
       struct http_context_s hd;
       
       err = http_open_document (&hd, url, 0);
       if (err)
-        log_error ("error retrieving `%s': %s\n", url, gpg_strerror (err));
+        log_error (_("error retrieving `%s': %s\n"), url, gpg_strerror (err));
       else if (hd.status_code != 200)
         {
-          log_error ("error retrieving `%s': http status %u\n",
+          log_error (_("error retrieving `%s': http status %u\n"),
                      url, hd.status_code);
           err = gpg_error (GPG_ERR_NO_DATA);
           http_close (&hd, 0);
@@ -134,13 +154,34 @@ crl_fetch (const char *url, FILE **r_fp)
     }
   else /* Let the LDAP code try other schemes. */
     {
-      err = url_fetch_ldap (url, "certificateRevocationList;binary",
-                            &value, &valuelen);
+      const char *s;
+      const char *attr;
+
+      /* FIXME: This is a bad hack.  I don't know why we at all
+         pass the attribute separately to the LDAP serach code.
+         The attributes are included in the URL and OpenLDAP
+         provides a parse functionality for it.  It has never been
+         used though.  We need to find out why this strange way
+         was used!!!!. */
+      s = strchr (url, '?');
+      if (s && !strcmp (s+1, "certificateRevocationList"))
+        attr = "certificateRevocationList;binary";
+      else if (s && !strcmp (s+1, "authorityRevocationList"))
+        attr = "authorityRevocationList;binary";
+      else 
+        {
+          log_info ("WARNING: unsupported LDAP attribute - using CRL\n");
+          attr = "certificateRevocationList;binary";
+        }
+
+      err = url_fetch_ldap (url, attr, &value, &valuelen);
       if (!err)
         err = setup_funopen (value, valuelen, r_fp);
-      else if (err && *value)
+      else if (err && value)
         xfree (value);
     }
+  if (free_this)
+    xfree (free_this);
   return err;
 }
 
@@ -151,35 +192,25 @@ gpg_error_t
 crl_fetch_default (const char *issuer, FILE **r_fp)
 {
   gpg_error_t err;
-  unsigned char *value = NULL;
-  size_t valuelen;
+/*   unsigned char *value = NULL; */
+/*   size_t valuelen; */
   
-  *r_fp = NULL;
+/*   *r_fp = NULL; */
   err = attr_fetch_ldap (issuer, "certificateRevocationList;binary", 
-                         &value, &valuelen);
-  if (!err)
-    err = setup_funopen (value, valuelen, r_fp);
-  else if (err && value)
-    xfree (value);
+                         r_fp);
+/*   if (!err) */
+/*     err = setup_funopen (value, valuelen, r_fp); */
+/*   else if (err && value) */
+/*     xfree (value); */
   return err;
 }
 
 
 /* Fetch a CA certificate for DN using the default server. */
 int
-ca_cert_fetch (const char *dn, unsigned char **value, size_t *valuelen)
+ca_cert_fetch (const char *dn, FILE **r_fp)
 {
-  int rc;
-
-  *value = NULL;
-  rc = attr_fetch_ldap (dn, "cACertificate;binary", value, valuelen);
-  if (rc && *value)
-    {
-      xfree (*value);
-      *value = NULL;
-      *valuelen = 0;
-    }
-  return rc;
+  return attr_fetch_ldap (dn, "cACertificate;binary", r_fp);
 }
 
 
