@@ -40,13 +40,9 @@
 #define JNLIB_NEED_LOG_LOGV
 #include "dirmngr.h"
 #include "crlcache.h"
+#include "crlfetch.h"
 #include "misc.h"
 
-/* no i18n needed */
-#define N_(a) a
-#define _(a) a
-
-
 enum cmd_and_opt_values 
 { aNull = 0,
   oCsh		  = 'c',
@@ -58,6 +54,7 @@ enum cmd_and_opt_values
   aServer,
   aListCRLs,
   aLoadCRL,
+  aFetchCRL,
   aShutdown,
   aFlush,
   aGPGConfList,
@@ -66,6 +63,7 @@ enum cmd_and_opt_values
   oDebug,
   oDebugAll,
   oDebugWait,
+  oDebugLevel,
   oNoGreeting,
   oNoOptions,
   oHomedir,
@@ -76,9 +74,11 @@ enum cmd_and_opt_values
   oLDAPTimeout,
   oLDAPAddServers,
   oOCSPResponder,
+  oOCSPSigner,
   oMaxReplies,
   oFakedSystemTime,
   oForce,
+  oAllowOCSP,
 aTest };
 
 
@@ -89,8 +89,9 @@ static ARGPARSE_OPTS opts[] = {
 
   { aServer,   "server",    256, N_("run in server mode") },
   { aListCRLs, "list-crls", 256, N_("list the contents of the CRL cache")},
-  { aLoadCRL,  "load-crl",  256, N_("load CRL from file into cache")},
-  { aShutdown, "shutdown",  256, N_("shutdown the agent")},
+  { aLoadCRL,  "load-crl",  256, N_("|FILE|load CRL from FILE into cache")},
+  { aFetchCRL, "fetch-crl", 256, N_("|URL|fetch a CRL from URL")},
+  { aShutdown, "shutdown",  256, N_("shutdown the dirmngr")},
   { aFlush,    "flush",     256, N_("flush the cache")},
   { aGPGConfList, "gpgconf-list", 256, "@" },
 
@@ -101,12 +102,12 @@ static ARGPARSE_OPTS opts[] = {
   { oSh,       "sh",        0, N_("sh-style command output") },
   { oCsh,      "csh",       0, N_("csh-style command output") },
   { oOptions,  "options"  , 2, N_("|FILE|read options from FILE")},
-  { oDebug,    "debug"     ,4|16, N_("|FLAGS|set the debugging FLAGS")},
-  { oDebugAll, "debug-all" ,0, N_("set all debugging flags")},
+  { oDebugLevel, "debug-level" ,2, N_("|NAME|set debugging level to NAME")},
   { oNoDetach, "no-detach" ,0, N_("do not detach from the console")},
   { oLogFile,  "log-file"  ,2, N_("|FILE|write logs to FILE")},
   { oBatch   , "batch"     ,0, N_("run without asking a user")},
   { oForce,    "force"     ,0, N_("force loading of outdated CRLs")},
+  { oAllowOCSP, "allow-ocsp",0,N_("allow sending OCSP requests")},
 
   { oLDAPFile, "ldapserverlist-file", 2,
     N_("|FILE|read LDAP server list from FILE")},
@@ -115,13 +116,15 @@ static ARGPARSE_OPTS opts[] = {
   { oLDAPTimeout, "ldaptimeout", 1,
     N_("|N|set LDAP timeout to N seconds")},
 
-  { oOCSPResponder, "ocsp-responder", 2, N_("|URL|Use OCSP Responder URL")},
-
+  { oOCSPResponder, "ocsp-responder", 2, N_("|URL|use OCSP responder at URL")},
+  { oOCSPSigner, "ocsp-signer", 2, N_("|FPR|OCSP response signed by FPR")}, 
 
   { oMaxReplies, "max-replies", 1,
     N_("|N|do not return more than N items in one query")},
 
   { oFakedSystemTime, "faked-system-time", 4, "@" }, /* (epoch time) */
+  { oDebug,    "debug"     ,4|16, "@"},
+  { oDebugAll, "debug-all" ,0,    "@"},
   { oDebugWait, "debug-wait", 1, "@"},
   { oNoGreeting, "no-greeting", 0, "@"},
   {0}
@@ -143,7 +146,7 @@ my_strusage( int level )
       case 11: p = "dirmngr";
 	break;
       case 13: p = VERSION; break;
-      case 14: p = "Copyright (C) 2002 Klarlvdalens Datakonsult AB"; break;
+      case 14: p = "Copyright (C) 2004 g10 Code GmbH"; break;
       case 17: p = PRINTABLE_OS_NAME; break;
       case 19: p =
 	    _("Please report bugs to <gpa-dev@gnupg.org>.\n");
@@ -154,7 +157,7 @@ my_strusage( int level )
 	break;
       case 41:	p =
 	    _("Syntax: dirmngr [options] [command [args]]\n"
-	      "LDAP access for GnuPG\n");
+	      "LDAP and OCSP access for GnuPG\n");
 	break;
 
       default:	p = NULL;
@@ -183,6 +186,87 @@ my_gcry_logger (void *dummy, int level, const char *fmt, va_list arg_ptr)
 }
 
 
+/* Callback from libksba to hash a provided buffer.  Our current
+   implementation does only allow SHA-1 for hashing. This may be
+   extended by mapping the name, testing for algorithm availibility
+   and adjust the length checks accordingly. */
+static gpg_error_t 
+my_ksba_hash_buffer (void *arg, const char *oid,
+                     const void *buffer, size_t length, size_t resultsize,
+                     unsigned char *result, size_t *resultlen)
+{
+  if (oid && strcmp (oid, "1.3.14.3.2.26")) 
+    return gpg_error (GPG_ERR_NOT_SUPPORTED); 
+  if (resultsize < 20)
+    return gpg_error (GPG_ERR_BUFFER_TOO_SHORT);
+  gcry_md_hash_buffer (2, result, buffer, length); 
+  *resultlen = 20;
+  return 0;
+}
+
+
+/* Setup the debugging.  With a LEVEL of NULL only the active debug
+   flags are propagated to the subsystems.  With LEVEL set, a specific
+   set of debug flags is set; thus overriding all flags already
+   set. */
+static void
+set_debug (const char *level)
+{
+  if (!level)
+    ;
+  else if (!strcmp (level, "none"))
+    opt.debug = 0;
+  else if (!strcmp (level, "basic"))
+    opt.debug = DBG_ASSUAN_VALUE;
+  else if (!strcmp (level, "advanced"))
+    opt.debug = DBG_ASSUAN_VALUE|DBG_X509_VALUE|DBG_LOOKUP_VALUE;
+  else if (!strcmp (level, "expert"))
+    opt.debug = (DBG_ASSUAN_VALUE|DBG_X509_VALUE|DBG_LOOKUP_VALUE
+                 |DBG_CACHE_VALUE|DBG_CRYPTO_VALUE);
+  else if (!strcmp (level, "guru"))
+    opt.debug = ~0;
+  else
+    {
+      log_error (_("invalid debug-level `%s' given\n"), level);
+      log_info (_("valid debug levels are: %s\n"),
+                "none, basic, advanced, expert, guru");
+      dirmngr_exit (2);
+    }
+
+
+  if (opt.debug && !opt.verbose)
+    {
+      opt.verbose = 1;
+      gcry_control (GCRYCTL_SET_VERBOSITY, (int)opt.verbose);
+    }
+  if (opt.debug && opt.quiet)
+    opt.quiet = 0;
+
+  if (opt.debug & DBG_CRYPTO_VALUE )
+    gcry_control (GCRYCTL_SET_DEBUG_FLAGS, 1);
+}
+ 
+
+static void
+i18n_init(void)
+{
+#ifdef USE_SIMPLE_GETTEXT
+  set_gettext_file (PACKAGE);
+#else
+# ifdef ENABLE_NLS
+#  ifdef HAVE_LC_MESSAGES
+  setlocale (LC_TIME, "");
+  setlocale (LC_MESSAGES, "");
+#  else
+  setlocale (LC_ALL, "" );
+#  endif
+  bindtextdomain (PACKAGE, LOCALEDIR);
+  textdomain (PACKAGE);
+# endif
+#endif
+}
+
+
 static void
 wrong_args (const char *text)
 {
@@ -209,6 +293,7 @@ main (int argc, char **argv )
   int default_config =1;
   int greeting = 0;
   int nogreeting = 0;
+  const char *debug_level = NULL;
   int nodetach = 0;
   int csh_style = 0;
   char *logfile = NULL;
@@ -226,7 +311,8 @@ main (int argc, char **argv )
   gcry_control (GCRYCTL_DISABLE_SECMEM, 0);
   if (!gcry_check_version (NEED_LIBGCRYPT_VERSION) )
     {
-      log_fatal (_("libgcrypt is too old (need %s, have %s)\n"),
+      log_fatal (_("%s is too old (need %s, have %s)\n"),
+                 "libgcrypt",
                  NEED_LIBGCRYPT_VERSION, gcry_check_version (NULL) );
     }
   gcry_set_log_handler (my_gcry_logger, NULL);
@@ -235,10 +321,12 @@ main (int argc, char **argv )
   /* Init KSBA.  */
   if (!ksba_check_version (NEED_KSBA_VERSION) )
     {
-      log_fatal( _("libksba is too old (need %s, have %s)\n"),
+      log_fatal( _("%s is too old (need %s, have %s)\n"),
+                 "libksba",
                  NEED_KSBA_VERSION, ksba_check_version (NULL) );
     }
   ksba_set_malloc_hooks (gcry_malloc, gcry_realloc, gcry_free );
+  ksba_set_hash_buffer_function (my_ksba_hash_buffer, NULL);
 
 
   /* Init Assuan. */
@@ -246,6 +334,8 @@ main (int argc, char **argv )
   assuan_set_assuan_log_stream (log_get_stream ());
   assuan_set_assuan_log_prefix (log_get_prefix (NULL));
 
+  /* Setup I18N. */
+  i18n_init();
 
   /* Setup defaults. */
   shell = getenv ("SHELL");
@@ -281,7 +371,7 @@ main (int argc, char **argv )
       if (pargs.r_opt == oDebug || pargs.r_opt == oDebugAll)
         parse_debug++;
       else if (pargs.r_opt == oOptions)
-        { /* yes there is one, so we do not try the default one, but
+        { /* Yes there is one, so we do not try the default one, but
 	     read the option file when it is encountered at the
 	     commandline */
           default_config = 0;
@@ -331,12 +421,12 @@ main (int argc, char **argv )
     {
       switch (pargs.r_opt)
         {
-
         case aServer: 
         case aShutdown: 
         case aFlush: 
 	case aListCRLs: 
 	case aLoadCRL: 
+        case aFetchCRL:
 	case aGPGConfList:
           cmd = pargs.r_opt;
           break;
@@ -347,10 +437,11 @@ main (int argc, char **argv )
 
         case oDebug: opt.debug |= pargs.r.ret_ulong; break;
         case oDebugAll: opt.debug = ~0; break;
+        case oDebugLevel: debug_level = pargs.r.ret_str; break;
         case oDebugWait: debug_wait = pargs.r.ret_int; break;
 
         case oOptions:
-          /* config files may not be nested (silently ignore them) */
+          /* Config files may not be nested (silently ignore them) */
           if (!configfp)
             {
 		xfree(configname);
@@ -374,9 +465,9 @@ main (int argc, char **argv )
 	  break;
 	case oMaxReplies: opt.max_replies = pargs.r.ret_int; break;
 
-        case oOCSPResponder:
-          add_to_strlist (&opt.ocsp_responders, pargs.r.ret_str);
-          break;
+        case oAllowOCSP: opt.allow_ocsp = 1; break;
+        case oOCSPResponder: opt.ocsp_responder = pargs.r.ret_str; break;
+        case oOCSPSigner:    opt.ocsp_signer = pargs.r.ret_str; break;
 
         case oFakedSystemTime:
           set_time ( (time_t)pargs.r.ret_ulong, 0);
@@ -417,9 +508,11 @@ main (int argc, char **argv )
     {
       dirmngr_isotime_t tbuf;
       get_isotime (tbuf);
-      log_info ("WARNING: running with faked system time: %s\n", tbuf);
+      log_info (_("WARNING: running with faked system time %s\n"), tbuf);
     }
 
+  set_debug (debug_level);
+
   /* Get LDAP server list from file. */
   if (!ldapfile) 
     {
@@ -430,9 +523,6 @@ main (int argc, char **argv )
   else
       opt.ldapservers = parse_ldapserver_file (ldapfile);
 
-  if (opt.ocsp_responders && opt.ocsp_responders->next)
-    log_info ("WARNING: only one OCSP responder currently supported\n");
-
   /* Ready.  Now to our duties. */
   if (!cmd)
     cmd = aServer;
@@ -440,7 +530,7 @@ main (int argc, char **argv )
   if (cmd == aServer)
     {
       if (argc)
-        wrong_args (_("--server"));
+        wrong_args ("--server");
 
       /* We need to ignore the PIPE signal because the we might log to
          a socket and that code handles EPIPE properly.  Assuan would
@@ -469,7 +559,7 @@ main (int argc, char **argv )
       crl_cache_t cache;
 
       if (argc)
-        wrong_args (_("--list-crls"));
+        wrong_args ("--list-crls");
       cache = crl_cache_init ();
       crl_cache_list (cache, stdout);
     }
@@ -486,11 +576,33 @@ main (int argc, char **argv )
             rc = crl_cache_load (NULL, c, *argv);
         }
     }
+  else if (cmd == aFetchCRL)
+    {
+      crl_cache_t c;
+      FILE *fp;
+
+      if (argc != 1)
+        wrong_args ("--fetch-crl URL");
+
+      c = crl_cache_init ();
+      rc = crl_fetch (argv[0], &fp);
+      if (rc)
+        log_error (_("fetching CRL from `%s' failed: %s\n"),
+                     argv[0], gpg_strerror (rc));
+      else
+        {
+          rc = crl_cache_insert (NULL, c,  argv[0], fp); 
+          if (rc)
+            log_error (_("processing CRL from `%s' failed: %s\n"),
+                       argv[0], gpg_strerror (rc));
+          fclose (fp);
+        }
+    }
   else if (cmd == aFlush)
     {
       /* Delete cache and exit. */
       if (argc)
-        wrong_args (_("--flush"));
+        wrong_args ("--flush");
       rc = crl_cache_flush();
     }
   else if (cmd == aGPGConfList)
@@ -516,12 +628,13 @@ main (int argc, char **argv )
       /* First the configuration file.  This is not an option, but it
 	 is vital information for GPG Conf.  */
       filename = make_filename (opt.homedir, "dirmngr.conf", NULL);
-      printf ("gpgconf-dirmngr.conf:%lu:\"%s\n", GC_OPT_FLAG_DEFAULT, filename);
+      printf ("gpgconf-dirmngr.conf:%lu:\"%s\n",
+              GC_OPT_FLAG_DEFAULT, filename);
       xfree (filename);
 
       printf ("verbose:%lu:\n", GC_OPT_FLAG_NONE);
       printf ("quiet:%lu:\n", GC_OPT_FLAG_NONE);
-      printf ("debug-all:%lu:\n", GC_OPT_FLAG_NONE);
+      printf ("debug-level:%lu:\"none\n", GC_OPT_FLAG_DEFAULT);
       printf ("log-file:%lu:\n", GC_OPT_FLAG_NONE);
       printf ("force:%lu:\n", GC_OPT_FLAG_NONE);
 
@@ -535,12 +648,15 @@ main (int argc, char **argv )
       printf ("ldapserverlist-file:%lu:\"%s\n", GC_OPT_FLAG_DEFAULT, filename);
       xfree (filename);
 
-      printf ("add-servers:%lu:\n", GC_OPT_FLAG_NONE);
-      printf ("ldaptimeout:%lu:%u\n", GC_OPT_FLAG_DEFAULT, DEFAULT_LDAP_TIMEOUT);
-      printf ("max-replies:%lu:%u\n", GC_OPT_FLAG_DEFAULT, DEFAULT_MAX_REPLIES);
+      printf ("ldaptimeout:%lu:%u\n",
+              GC_OPT_FLAG_DEFAULT, DEFAULT_LDAP_TIMEOUT);
+      printf ("max-replies:%lu:%u\n",
+              GC_OPT_FLAG_DEFAULT, DEFAULT_MAX_REPLIES);
+      printf ("allow-ocsp:%lu:\n", GC_OPT_FLAG_NONE);
+      printf ("ocsp-responder:%lu:\n", GC_OPT_FLAG_NONE);
+      printf ("ocsp-signer:%lu:\n", GC_OPT_FLAG_NONE);
 
       printf ("faked-system-time:%lu:\n", GC_OPT_FLAG_NONE);
-      printf ("debug-wait:%lu:\n", GC_OPT_FLAG_NONE);
       printf ("no-greeting:%lu:\n", GC_OPT_FLAG_NONE);
     }
   cleanup ();
@@ -595,13 +711,11 @@ parse_ldapserver_file (const char* filename)
   unsigned int lineno = 0;
   FILE *fp;
   int fieldno;
-  int fail = 0;
-  int errcount = log_get_errorcount (0);
 
   fp = fopen (filename, "r");
   if (!fp)
     {
-      log_error ("error opening `%s': %s\n", filename, strerror (errno));
+      log_error (_("error opening `%s': %s\n"), filename, strerror (errno));
       return NULL;
     }
 
@@ -609,14 +723,21 @@ parse_ldapserver_file (const char* filename)
   serverend = &serverstart;
   while (fgets (buffer, sizeof buffer, fp))
     {
+      int fail = 0;
+
       lineno++;
       if (!*buffer || buffer[strlen(buffer)-1] != '\n')
         {
-          log_error (_("%s:%u: line too long\n"), filename, lineno);
-          fail = 1;
-          while ( (c=fgetc (fp)) != EOF && c != '\n')
-            ; /* Skip until end of line. */
-          continue;
+          if (*buffer && feof (fp))
+            ; /* Last line not terminated - continue. */
+          else
+            {
+              log_error (_("%s:%u: line too long - skipped\n"),
+                         filename, lineno);
+              while ( (c=fgetc (fp)) != EOF && c != '\n')
+                ; /* Skip until end of line. */
+              continue;
+            }
         }
       /* Skip empty and comment lines.*/
       for (p=buffer; spacep (p); p++)
@@ -638,7 +759,11 @@ parse_ldapserver_file (const char* filename)
               if (*p)
                 server->host = xstrdup (p);
               else
-                log_error (_("%s:%u: no hostname given\n"), filename, lineno);
+                {
+                  log_error (_("%s:%u: no hostname given\n"),
+                             filename, lineno);
+                  fail = 1;
+                }
               break;
                 
             case 2:
@@ -653,8 +778,11 @@ parse_ldapserver_file (const char* filename)
 
             case 4:
               if (*p && !server->user)
-                log_error (_("%s:%u: password given without user\n"), 
-                           filename, lineno);
+                {
+                  log_error (_("%s:%u: password given without user\n"), 
+                             filename, lineno);
+                  fail = 1;
+                }
               else if (*p)
                 server->pass = xstrdup (p);
               break;
@@ -670,21 +798,22 @@ parse_ldapserver_file (const char* filename)
             }
         }
           
-      *serverend = server;
-      serverend = &server->next;
+      if (fail)
+        {
+          log_info (_("%s:%u: skipping this line\n"), filename, lineno);
+          free_ldapservers_list (server);
+        }
+      else
+        {
+          *serverend = server;
+          serverend = &server->next;
+        }
     } 
   
   if (ferror (fp))
     log_error (_("error reading `%s': %s\n"), filename, strerror (errno));
   fclose (fp);
 
-  if (log_get_errorcount (0) > errcount)
-    {
-      free_ldapservers_list (serverstart);
-      serverstart = NULL;
-      log_error (_("errors found in `%s' - file ignored\n"), filename);
-    }
-
   return serverstart;
 }
 
