@@ -464,7 +464,8 @@ send_request (http_t hd)
       err = http_parse_uri (&uri, http_proxy);
       if (err)
         {
-          log_error ("invalid " HTTP_PROXY_ENV_PRINTABLE ": %s\n",
+          log_error (_("invalid %s: %s\n"),
+                     HTTP_PROXY_ENV_PRINTABLE,
                      gpg_strerror (err));
           http_release_parsed_uri (uri);
           return gpg_error (GPG_ERR_CONFIGURATION);
@@ -816,7 +817,7 @@ connect_server (const char *server, unsigned short port)
   hints.ai_socktype = SOCK_STREAM;
   if (getaddrinfo (server, portstr, &hints, &res))
     {
-      log_error ("error resolving `%s': host not found\n", server);
+      log_error (_("error resolving `%s': host not found\n"), server);
       errno = ENOENT; /* fixme: Is there a better one? */
       return -1;
     }
@@ -829,7 +830,7 @@ connect_server (const char *server, unsigned short port)
       if (sock == -1)
         {
           int save_errno = errno;
-          log_error ("error creating socket: %s\n", strerror (errno));
+          log_error (_("error creating socket: %s\n"), strerror (errno));
           freeaddrinfo (res);
           errno = save_errno;
           return -1;
@@ -852,7 +853,7 @@ connect_server (const char *server, unsigned short port)
   host = gethostbyname (server); 
   if (!host)
     {
-      log_error ("error resolving `%s': host not found\n", server);
+      log_error (_("error resolving `%s': host not found\n"), server);
       errno = ENOENT; /* fixme: Is there a better one? */
       return -1;
     }
@@ -860,7 +861,7 @@ connect_server (const char *server, unsigned short port)
   sock = socket (host->h_addrtype, SOCK_STREAM, 0);
   if (sock == -1)
     {
-      log_error ("error creating socket: %s\n", strerror (errno));
+      log_error (_("error creating socket: %s\n"), strerror (errno));
       return -1;
     }
 
@@ -915,7 +916,7 @@ write_server (int sock, const char *data, size_t length)
               select (0, NULL, NULL, NULL, &tv);
               continue;
             }
-          log_info ("network write failed: %s\n", strerror (errno));
+          log_info (_("network write failed: %s\n"), strerror (errno));
           return gpg_error_from_errno (errno);
         }
       nleft -= nwritten;
