@@ -29,6 +29,7 @@
 
 #include "crlfetch.h"
 #include "dirmngr.h"
+#include "misc.h"
 
 
 #define CERTATTR "userCertificate"
@@ -47,6 +48,7 @@ struct cert_fetch_context_s {
 };
 
 
+
 /* Fetch the value attr, put result in value, length in valuelen.
    Returns 0 on success or an gpg error code.
  */
@@ -67,7 +69,7 @@ get_attr_from_result_ldap (LDAP * ld, LDAPMessage * msg, const char *attr,
   values = ldap_get_values_len (ld, item, attr);
 
   if (values && opt.verbose)
-    log_info ("found attribute `%s'\n", attr);
+    log_info (_("found attribute `%s'\n"), attr);
 
   if (!values)
     {
@@ -80,7 +82,7 @@ get_attr_from_result_ldap (LDAP * ld, LDAPMessage * msg, const char *attr,
       if (s)
         { 
           newattr = xstrdup (attr);
-          newattr[s - newattr] = 0;
+          newattr[s - attr] = 0;
         }
       else
         {
@@ -90,7 +92,7 @@ get_attr_from_result_ldap (LDAP * ld, LDAPMessage * msg, const char *attr,
 
       values = ldap_get_values_len (ld, item, newattr);
       if (values && opt.verbose)
-	    log_info ("found attributes `%s' at second try\n", newattr);
+	    log_info (_("found attributes `%s' at second try\n"), newattr);
       xfree (newattr);
     }
 
@@ -106,21 +108,226 @@ get_attr_from_result_ldap (LDAP * ld, LDAPMessage * msg, const char *attr,
     return gpg_error (GPG_ERR_NO_CRL_KNOWN);
 }
 
+/* The cookie structure used for returning the results from a
+   attr_fetch_ldap query. */
+struct attr_fetch_fun_cookie_s {
+  int eof_reached;   /* True if we already encountered EOF. */
+  gpg_error_t err;   /* The error code from the last operation. */
+  int sys_err;       /* The system error number. */
+  char *attr;        /* The attribute used for the search. */
+  LDAP *ld;
+  int msgid;
+  LDAPMessage *result;
+  LDAPMessage *item;
+  struct berval **values;
+  int nread;
+};
+
+
+/* The reader function used for returning the results from a
+   attr_fetch_ldap query. */
+static int 
+attr_fetch_fun_reader (void *cookie, char *buffer, int size)
+{
+  struct attr_fetch_fun_cookie_s *h = cookie;
+  int rc;
+  struct timeval timeout, tout;
+
+  if (h->eof_reached || !size)
+    return 0;
+
+  dirmngr_tick (NULL);
+  if (h->err)
+    {
+      errno = h->sys_err;
+      return -1;
+    }
+
+  for (;;)
+    {
+      /* Any pending stuff we need to return first? */
+      if (h->values && h->values[0] && h->values[0]->bv_len)
+        {
+          if (h->nread < h->values[0]->bv_len)
+            {
+              if (h->nread + size > h->values[0]->bv_len)
+                size = h->values[0]->bv_len - h->nread;
+              memcpy (buffer, h->values[0]->bv_val + h->nread, size);
+              h->nread += size;
+              return size;
+            }
+          if (h->values)
+            {
+              ldap_value_free_len (h->values);
+              h->values = NULL;
+            }
+        }
+
+      /* If we still have a result from the last round, process that one. */
+      if (h->result)
+        {
+          if (!h->item)
+            h->item = ldap_first_entry (h->ld, h->result);
+          else
+            h->item = ldap_next_entry (h->ld, h->item);
+        }
+      
+      /* Continue if we still hold an item. */
+      if (h->result && h->item)
+        {
+          if (h->values)
+            ldap_value_free_len (h->values);
+          h->values = ldap_get_values_len (h->ld, h->item, h->attr);
+
+          if (h->values && opt.verbose)
+            log_info (_("found attribute `%s'\n"), h->attr);
+
+          if (!h->values)
+            {
+              /* Try again without or with ";binary" .. which worked with some
+                 strange LDAP servers.  */
+              const char *s;
+              char *newattr;
+
+              s = memistr (h->attr, strlen (h->attr), ";binary");
+              if (s)
+                { 
+                  newattr = xstrdup (h->attr);
+                  newattr[s - newattr] = 0;
+                }
+              else
+                {
+                  newattr = xmalloc (strlen (h->attr) + 7 + 1);
+                  strcpy (stpcpy (newattr, h->attr), ";binary");
+                }
+
+              h->values = ldap_get_values_len (h->ld, h->item, newattr);
+              if (h->values && opt.verbose)
+                log_info (_("found attributes `%s' at second try\n"), newattr);
+              xfree (newattr);
+            }
+        }
+
+      /* If there is stuff to return, jump to the top.  Note, that we
+         only handle the first item in the list of values. */
+      if (h->values && h->values[0] && h->values[0]->bv_len)
+        continue;
+
+      /* Wait for the next result. */
+      timeout.tv_sec = 0;
+      timeout.tv_usec = 0;
+    retry:
+      tout.tv_sec = 0;
+      tout.tv_usec = 100000; /* Temporary timeout after 100 ms. */
+      if (h->result)
+        {
+          ldap_msgfree (h->result);
+          h->result = NULL;
+        }
+      rc = ldap_result (h->ld, h->msgid, 0, &tout, &h->result);
+      if (!rc)
+        {
+          /* We don't take an updated timeout into account, not all
+             systems do this properly in select and we don't know whether
+             ldap_result fixes for that. We also ignore the usec which is
+             not a problem as we set opt.ldaptimeout only to full seconds.
+             The timeout is restarted with every bit of data we get. */
+          timeout.tv_usec += 100000;
+          while (timeout.tv_usec >= 1000000)
+            {
+              timeout.tv_sec++;
+              timeout.tv_usec -= 1000000;
+            }
+          if (timeout.tv_sec < opt.ldaptimeout.tv_sec)
+            {
+              dirmngr_tick (NULL);
+              goto retry;
+            }
+
+          log_error (_("timeout while waiting for ldap_search result\n"));
+          h->err = gpg_error (GPG_ERR_TIMEOUT);
+          h->sys_err = errno = ETIMEDOUT;
+          return -1;
+        }
+      if (rc == -1)
+        {
+          log_error (_("ldap_result for an ldap_search failed: %s\n"),
+                     strerror (errno));
+          h->sys_err = errno;
+          h->err = gpg_error (GPG_ERR_NO_CRL_KNOWN);
+          return -1;    
+        }
+      else if (rc == LDAP_RES_SEARCH_ENTRY)
+        ; /* Yep, we got a result. */
+      else if (rc == LDAP_RES_SEARCH_RESULT)
+        {
+          int errcode;
+
+          rc = ldap_parse_result (h->ld, h->result, &errcode, NULL, NULL,
+                                  NULL, NULL, 0);
+          if (rc)
+            log_error (_("ldap_parse_result failed\n"));
+          else 
+            {
+              if (errcode)
+                log_info (_("ldap search result is: %s\n"),
+                          ldap_err2string (errcode));
+            }
+          h->eof_reached = 1;
+          return 0;
+        }
+      else
+        {
+          log_error (_("unexpected message type %d "
+                       "returned for ldap_search\n"), rc);
+          h->sys_err = errno = 0;
+          h->err = gpg_error (GPG_ERR_NO_CRL_KNOWN);
+          return -1;    
+        }
+    }
+  /*NOT REACHED*/
+}
+
+/* The closer function used for returning the results from a
+   attr_fetch_ldap query. */
+static int
+attr_fetch_fun_closer (void *cookie)
+{
+  struct attr_fetch_fun_cookie_s *h = cookie;
+  
+  if (h->values)
+    ldap_value_free_len (h->values);
+  if (h->result)
+    {
+      ldap_msgfree (h->result);
+      h->result = NULL;
+    }
+  xfree (h->attr);
+  ldap_unbind (h->ld);
+  xfree (h);
+  return 0;
+}
+
+
 
 /* Perform an LDAP query.  Returns an gpg error code or 0 on success.
-   The values are returned in VALUE and VALUELEN. */
+   The function returns a new stream at R_FP. */
 static gpg_error_t
 attr_fetch_ldap_internal (const char *host, int port,
 			  const char *user, const char *pass,
 			  const char *dn, const char *attr,
-			  unsigned char **value, size_t * valuelen)
+			  FILE **r_fp)
 {
   gpg_error_t err;
+  struct timeval timeout;
   char *attrs[2];
   int rc;
   int msgid;
   LDAPMessage *result;
   LDAP *ld;
+  struct attr_fetch_fun_cookie_s *cookie;
+
+  *r_fp = NULL;
 
   attrs[0] = xstrdup (attr); 
   attrs[1] = NULL;
@@ -132,7 +339,7 @@ attr_fetch_ldap_internal (const char *host, int port,
   ld = ldap_init (host, port);
   if (!ld)
     {
-      log_error ("ldap_init failed for '%s:%d': %s\n",
+      log_error (_("ldap_init failed for '%s:%d': %s\n"),
                  host, port, strerror (errno));
       /* We do not return the system error because the host might
          simply not be available and the user is not intersted inthese
@@ -142,8 +349,8 @@ attr_fetch_ldap_internal (const char *host, int port,
   msgid = ldap_simple_bind (ld, user, pass);
   if (msgid == -1)
     {
-      log_error ("ldap_simple_bind failed for user '%s': %s\n",
-                 user? user : "[none]", strerror (errno));
+      log_error (_("ldap_simple_bind failed for user '%s': %s\n"),
+                 user? user : _("[none]"), strerror (errno));
       /* According to the documentation a variable ld->ld_errno will
          be set.  However, that structure is private and thus we can't
          access.  So we assume that the standard errno is used. */
@@ -152,16 +359,18 @@ attr_fetch_ldap_internal (const char *host, int port,
       return gpg_error (GPG_ERR_NO_CRL_KNOWN);
     }
 
-  rc = ldap_result (ld, msgid, 0, &opt.ldaptimeout, &result);
+  dirmngr_tick (NULL);
+  timeout = opt.ldaptimeout;
+  rc = ldap_result (ld, msgid, 0, &timeout, &result);
   if (rc != LDAP_RES_BIND)
     {
       if (rc == -1)
-        log_error ("ldap_result for bind failed: %s\n", strerror (errno));
+        log_error (_("ldap_result for bind failed: %s\n"), strerror (errno));
       else if (!rc)
-        log_error ("timeout while waiting for ldap_bind to complete\n");
+        log_error (_("timeout while waiting for ldap_bind to complete\n"));
       else
         {
-          log_error ("unexpected message type %d returned for ldap_bind\n",
+          log_error (_("unexpected message type %d returned for ldap_bind\n"),
                      rc);
           ldap_msgfree (result);
         }
@@ -172,29 +381,47 @@ attr_fetch_ldap_internal (const char *host, int port,
   ldap_msgfree (result);
   result = NULL;
 
-  rc = ldap_search_st (ld, dn, LDAP_SCOPE_BASE, "objectclass=*", attrs, 0,
-                       &opt.ldaptimeout, &result);
-  if (rc)
+  /* Start the search. */
+  dirmngr_tick (NULL);
+  msgid = ldap_search (ld, dn, LDAP_SCOPE_BASE, "objectclass=*", attrs, 0);
+  if (msgid == -1)
     {
-      log_error ("ldap_search failed: %s\n", ldap_err2string (rc));
+      log_error (_("ldap_search failed: %s\n"), strerror (errno));
       xfree (attrs[0]);
       ldap_unbind (ld);
       return gpg_error (GPG_ERR_NO_CRL_KNOWN);
     }
 
-  err = get_attr_from_result_ldap (ld, result, attr, value, valuelen);
+  /* Setup the stream. */
+  cookie = xtrycalloc (1, sizeof *cookie);
+  if (!cookie)
+    {
+      err = gpg_error_from_errno (errno);
+      xfree (attrs[0]);
+      ldap_unbind (ld);
+      return err;
+    }
 
-  ldap_msgfree (result);
-  xfree (attrs[0]);
-  return err;
+  cookie->ld = ld;
+  cookie->msgid = msgid;
+  cookie->attr  = attrs[0];
+  *r_fp = funopen (cookie, attr_fetch_fun_reader, NULL, NULL, 
+                   attr_fetch_fun_closer);
+  if (!*r_fp)
+    {
+      err = gpg_error_from_errno (errno);
+      attr_fetch_fun_closer (cookie);
+      return err;
+    }
+
+  return 0;
 }
 
 
 /* Perform an LDAP query on all configured servers.  On error the
-   error code of the last try is retruned.  */
+   error code of the last try is returned.  */
 gpg_error_t
-attr_fetch_ldap (const char *dn, const char *attr,
-                 unsigned char **value, size_t * valuelen)
+attr_fetch_ldap (const char *dn, const char *attr, FILE **r_fp)
 {
   struct ldap_server_s *server;
   gpg_error_t err = gpg_error (GPG_ERR_CONFIGURATION);
@@ -203,9 +430,9 @@ attr_fetch_ldap (const char *dn, const char *attr,
     {
       err = attr_fetch_ldap_internal (server->host, server->port,
 				     server->user, server->pass,
-				     dn, attr, value, valuelen);
+				     dn, attr, r_fp);
       if (!err)
-        break; /* Found a result. Ready. */
+        break; /* Probably found a result. Ready. */
     }
   return err;
 }
@@ -224,14 +451,16 @@ url_fetch_ldap_internal (const LDAPURLDesc *ludp, const char *attr,
   ld = ldap_init (ludp->lud_host, ludp->lud_port);
   if (!ld)
     {
-      log_error ("ldap_init failed for '");
+      log_error (_("%s failed for `"), "ldap_init");
       print_sanitized_string (log_get_stream (), ludp->lud_host, 0);
       log_printf (":%d': %s\n", ludp->lud_port, strerror (errno));
       return gpg_error (GPG_ERR_NO_CRL_KNOWN);
     }
   if (ldap_simple_bind_s (ld, 0, 0))
     {
-      log_error ("ldap_simple_bind failed: %s\n", strerror (errno));
+      log_error (_("%s failed for `"), "ldap_simple_bind");
+      print_sanitized_string (log_get_stream (), ludp->lud_host, 0);
+      log_printf (":%d': %s\n", ludp->lud_port, strerror (errno));
       return gpg_error (GPG_ERR_NO_CRL_KNOWN);
     }
 
@@ -240,7 +469,9 @@ url_fetch_ldap_internal (const LDAPURLDesc *ludp, const char *attr,
                        &opt.ldaptimeout, &result);
   if (rc)
     {
-      log_error ("ldap_search failed: %s\n", ldap_err2string (rc));
+      log_error (_("%s failed for `"), "ldap_search");
+      print_sanitized_string (log_get_stream (), ludp->lud_host, 0);
+      log_printf (":%d': %s\n", ludp->lud_port, strerror (errno));
       return gpg_error (GPG_ERR_NO_CRL_KNOWN);
     }
 
@@ -276,12 +507,12 @@ add_server_to_servers (const char *host, int port)
                  "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                  "01234567890.-", *s))
       {
-        log_error ("invalid char 0x%02x in host name"
-                   " - not added", *s);
+        log_error (_("invalid char 0x%02x in host name"
+                   " - not added\n"), *s);
         return;
       }
 
-  log_info ("adding `%s:%d' to the ldap server list\n", host, port);
+  log_info (_("adding `%s:%d' to the ldap server list\n"), host, port);
   server = xcalloc (1, sizeof *s);
   server->host = xstrdup (host);
   server->port = port;
@@ -313,7 +544,7 @@ url_fetch_ldap (const char *url, const char *attr, unsigned char **value,
     {
       log_error ("'");
       print_sanitized_string (log_get_stream (), url, 0);
-      log_printf ("' is not an LDAP URL\n");
+      log_printf (_("' is not an LDAP URL\n"));
       return gpg_error (GPG_ERR_INV_URI);
     }
 
@@ -321,7 +552,7 @@ url_fetch_ldap (const char *url, const char *attr, unsigned char **value,
     {
       log_error ("'");
       print_sanitized_string (log_get_stream (), url, 0);
-      log_printf ("' is an invalid LDAP URL\n");
+      log_printf (_("' is an invalid LDAP URL\n"));
       return gpg_error (GPG_ERR_BAD_URI);
     }
 
@@ -499,7 +730,8 @@ start_cert_fetch_ldap (cert_fetch_context_t *context,
       sl = parse_one_pattern (patterns->d);
       if (!sl)
         {
-          log_error ("start_cert_fetch: invalid pattern `%s'\n", patterns->d);
+          log_error (_("start_cert_fetch: invalid pattern `%s'\n"),
+                     patterns->d);
           free_strlist (query_list);
           return gpg_error (GPG_ERR_INV_USER_ID);
         }
@@ -512,8 +744,9 @@ start_cert_fetch_ldap (cert_fetch_context_t *context,
   if (!ld)
     {
       err = gpg_error_from_errno (errno);
-      log_error ("ldap_init failed for '%s:%d': %s\n",
-                 host, port, strerror (errno));
+      log_error (_("%s failed for `"), "ldap_init");
+      print_sanitized_string (log_get_stream (), host, 0);
+      log_printf (":%d': %s\n", port, strerror (errno));
       free_strlist (query_list);
       return err;
     }
@@ -522,8 +755,10 @@ start_cert_fetch_ldap (cert_fetch_context_t *context,
   if (msgid == -1)
     {
       err = gpg_error_from_errno (errno);
-      log_error ("ldap_simple_bind failed for user '%s': %s\n",
-                 user? user : "[none]", strerror (errno));
+      log_error (_("%s failed for `"), "ldap_simple_bind");
+      print_sanitized_string (log_get_stream (), host, 0);
+      log_printf (":%d' user `%s': %s\n",
+                  port, user? user : _("[none]"),strerror (errno));
       ldap_unbind (ld);
       free_strlist (query_list);
       return err;
@@ -532,19 +767,18 @@ start_cert_fetch_ldap (cert_fetch_context_t *context,
   rc = ldap_result (ld, msgid, 0, &opt.ldaptimeout, &result);
   if (rc != LDAP_RES_BIND)
     {
-      if (rc == -1)
+      if (rc == -1 || !rc)
         {
-          err = gpg_error_from_errno (errno);
-          log_error ("ldap_result for bind failed: %s\n", strerror (errno));
-        }
-      else if (!rc)
-        {
-          err = gpg_error (GPG_ERR_TIMEOUT);
-          log_error ("timeout while waiting for ldap_bind to complete\n");
+          err = rc? gpg_error_from_errno (errno) : gpg_error (GPG_ERR_TIMEOUT);
+          log_error (_("%s failed for `"), "ldap_result/bind");
+          print_sanitized_string (log_get_stream (), host, 0);
+          log_printf (":%d' user `%s': %s\n",
+                      port, user? user : _("[none]"), gpg_strerror (err));
         }
       else
         {
-          log_error ("unexpected message type %d returned for ldap_bind\n", rc);
+          log_error (_("unexpected message type %d returned for ldap_bind\n"),
+                     rc);
           ldap_msgfree (result);
           err = gpg_error (GPG_ERR_BUG);
         }
@@ -682,7 +916,7 @@ fetch_next_cert_ldap (cert_fetch_context_t context,
           if (rc == LDAP_SIZELIMIT_EXCEEDED)
             {
               truncated = 1;
-              log_info ("ldap_search hit the size limit of the server\n");
+              log_info (_("ldap_search hit the size limit of the server\n"));
               /* Fixme: Shouldn't we bail out here? */
             }
           else if (rc == LDAP_NO_SUCH_OBJECT)
@@ -694,7 +928,7 @@ fetch_next_cert_ldap (cert_fetch_context_t context,
             }
           else
             {
-              log_error ("ldap_search failed: %s\n",
+              log_error (_("ldap_search failed: %s\n"),
                          ldap_err2string (ldap_result2error
                                           (context->ld,
                                            context->result, 0)));
@@ -734,6 +968,6 @@ end_cert_fetch_ldap (cert_fetch_context_t context)
       ldap_unbind (context->ld);
       xfree (context);
 
-      /* fixme: do we need to release the BERVALUES. */
+      /* fixme: do we need to release the BERVALUES? */
     }
 }
