@@ -1,6 +1,6 @@
 /* dirmngr.c - LDAP access
  *	Copyright (C) 2002 Klarlvdalens Datakonsult AB
- *      Copyright (C) 2003 g10 Code GmbH
+ *      Copyright (C) 2003, 2004 g10 Code GmbH
  *
  * This file is part of DirMngr.
  *
@@ -38,14 +38,13 @@
 #include "dirmngr.h"
 #include "ocsp.h"
 
-/* no i18n needed */
-#define N_(a) a
-#define _(a) a
-
-
 #define PARM_ERROR(t) assuan_set_error (ctx, ASSUAN_Parameter_Error, (t))
 
 
+/* We keep track of the primary client using us.  This is mainly a
+   hack to send the progress info back. */
+static ctrl_t primary_connection;
+
 /* Data used to associate an Assuan context with local server data */
 struct server_local_s {
   ASSUAN_CONTEXT assuan_ctx;
@@ -100,7 +99,7 @@ map_assuan_err (int err)
   return gpg_err_make (GPG_ERR_SOURCE_UNKNOWN, ec);
 }
 
-/* Map GPG_xERR_xx error codes to Assuan status codes */
+/* Map GPG_ERR_xx error codes to Assuan status codes */
 static int
 map_to_assuan_status (int rc)
 {
@@ -152,32 +151,23 @@ strcpy_escaped_plus (char *d, const unsigned char *s)
 }
 
 
-
-/* Ask back to return a certificate for ISSUER, given as a regular
-   gpgsm certifciate indentidicates (e.g. fingerprint or one of the
-   other methods).  Either return the certificate in a KSBA object or
-   NULL if it is not available. 
-
-   Note: get_issuer_cert_loal is a misnomer as the function allows us
-   to get back any certificate.
-*/
-ksba_cert_t 
-get_issuer_cert_local (ctrl_t ctrl, const char *issuer)
+/* Common code for get_cert_local and get_issuer_cert_local. */
+static ksba_cert_t 
+do_get_cert_local (ctrl_t ctrl, const char *name, const char *command)
 {
   unsigned char *value;
   size_t valuelen; 
   int rc;
   char *buf;
-  ksba_cert_t issuer_cert;
+  ksba_cert_t cert;
 
-  if (!ctrl || !ctrl->server_local->assuan_ctx)
+  if (name)
     {
-      log_debug ("get_issuer_cert_local called w/o context\n");
-      return NULL;
+      buf = xmalloc ( strlen (command) + 1 + strlen(name) + 1);
+      strcpy (stpcpy (stpcpy (buf, command), " "), name);
     }
-
-  buf = xmalloc ( 9 + strlen(issuer) + 1);
-  strcpy (stpcpy (buf, "SENDCERT "), issuer);
+  else
+    buf = xstrdup (command);
 
   /* FIXME: When changed to run as a system wide server, we should
      setup a maxlen limit for the certificate.  The problem is that
@@ -194,31 +184,64 @@ get_issuer_cert_local (ctrl_t ctrl, const char *issuer)
       return NULL;
     }
   
-  rc = ksba_cert_new (&issuer_cert);
+  rc = ksba_cert_new (&cert);
   if (!rc)
     {
-      rc = ksba_cert_init_from_mem (issuer_cert, value, valuelen);
+      rc = ksba_cert_init_from_mem (cert, value, valuelen);
       if (rc)
         {
-          ksba_cert_release (issuer_cert);
-          issuer_cert = NULL;
+          ksba_cert_release (cert);
+          cert = NULL;
         }
     }
   xfree (value);
-  return issuer_cert;
+  return cert;
+}
+
+
+
+/* Ask back to return a certificate for name, given as a regular
+   gpgsm certificate indentificates (e.g. fingerprint or one of the
+   other methods).  Alternatively, NULL may be used for NAME to
+   return the current target certificate. Either return the certificate
+   in a KSBA object or NULL if it is not available.
+*/
+ksba_cert_t 
+get_cert_local (ctrl_t ctrl, const char *name)
+{
+  if (!ctrl || !ctrl->server_local->assuan_ctx)
+    {
+      log_debug ("get_cert_local called w/o context\n");
+      return NULL;
+    }
+  return do_get_cert_local (ctrl, name, "SENDCERT");
+
+}
+
+/* Ask back to return the issuing certificate for name, given as a
+   regular gpgsm certificate indentificates (e.g. fingerprint or one
+   of the other methods).  Alternatively, NULL may be used for NAME to
+   return thecurrent target certificate. Either return the certificate
+   in a KSBA object or NULL if it is not available.
+*/
+ksba_cert_t 
+get_issuing_cert_local (ctrl_t ctrl, const char *name)
+{
+  if (!ctrl || !ctrl->server_local->assuan_ctx)
+    {
+      log_debug ("get_issuing_cert_local called w/o context\n");
+      return NULL;
+    }
+  return do_get_cert_local (ctrl, name, "SENDISSUERCERT");
 }
 
 
 /* Ask the client to return the certificate asscociated with the
-   current command. This is sometime needed because the client usually
+   current command. This is sometimes needed because the client usually
    sends us just the cert ID, assuming that the request can be
    satisfied from the cache, where the cert ID is used as key. */
-
-/* FIXME: We need to check the logic and better rename the function
-   becuase it does more than merely get a certificate - it is actually
-   the core of the the CRL retrievel. */
 static int
-inquire_cert (assuan_context_t ctx)
+inquire_cert_and_load_crl (assuan_context_t ctx)
 {
   ctrl_t ctrl = assuan_get_pointer (ctx);
   gpg_error_t err;
@@ -292,7 +315,7 @@ inquire_cert (assuan_context_t ctx)
       err = crl_fetch (distpoint_uri, &crl_fp);
       if (err)
         {
-          log_error ("crl_fetch via DP failed: %s\n", gpg_strerror (err));
+          log_error (_("crl_fetch via DP failed: %s\n"), gpg_strerror (err));
           goto leave;
         }
       
@@ -300,7 +323,7 @@ inquire_cert (assuan_context_t ctx)
                               distpoint_uri, crl_fp); 
       if (err)
         {
-          log_error ("crl_cache_insert via DP failed: %s\n",
+          log_error (_("crl_cache_insert via DP failed: %s\n"),
                      gpg_strerror (err));
           goto leave;
         }
@@ -323,7 +346,8 @@ inquire_cert (assuan_context_t ctx)
       err = crl_fetch_default (issuer, &crl_fp);
       if (err)
           {
-            log_error ("crl_fetch via issuer failed: %s\n", gpg_strerror (err));
+            log_error (_("crl_fetch via issuer failed: %s\n"),
+                       gpg_strerror (err));
             goto leave;
           }
 
@@ -331,7 +355,7 @@ inquire_cert (assuan_context_t ctx)
                              "default location(s)", crl_fp);
       if (err)
         {
-          log_error ("crl_cache_insert via issuer failed: %s\n",
+          log_error (_("crl_cache_insert via issuer failed: %s\n"),
                      gpg_strerror (err));
           goto leave;
         }
@@ -351,6 +375,23 @@ inquire_cert (assuan_context_t ctx)
 }
 
 
+/* Handle OPTION commands. */
+static int
+option_handler (ASSUAN_CONTEXT ctx, const char *key, const char *value)
+{
+  ctrl_t ctrl = assuan_get_pointer (ctx);
+
+  if (!strcmp (key, "force-crl-refresh"))
+    {
+      int i = *value? atoi (value) : 0;
+      ctrl->force_crl_refresh = i;
+    }
+  else
+    return ASSUAN_Invalid_Option;
+
+  return 0;
+}
+
 
 /* IS_VALID <certificate_id>|<certificate_fpr>
   
@@ -395,7 +436,7 @@ cmd_isvalid (assuan_context_t ctx, char *line)
       if (strlen (issuerhash) != 40)
         {
           xfree (issuerhash);
-          return PARM_ERROR ("serialno missing in cert ID");
+          return PARM_ERROR (_("serialno missing in cert ID"));
         }
       ocsp_mode = 1;
     }
@@ -404,9 +445,13 @@ cmd_isvalid (assuan_context_t ctx, char *line)
  again:
   if (ocsp_mode)
     {
-      char *cert_fpr = issuerhash;
-
-      err = ocsp_isvalid (ctrl, cert_fpr);
+      /* Note, that we ignore the given issuer hash and instead rely
+         on the current certificate semantics used with this
+         command. */
+      if (!opt.allow_ocsp)
+        err = gpg_error (GPG_ERR_NOT_SUPPORTED);
+      else
+        err = ocsp_isvalid (ctrl, NULL);
       /* Fixme: If we got no ocsp response we should fall back to CRL
          mode.  Thus we need to clear OCSP_MODE, get the issuerhash
          and the serialno from the current certificate and jump to
@@ -415,7 +460,8 @@ cmd_isvalid (assuan_context_t ctx, char *line)
   else
     {
       switch (crl_cache_isvalid (ctrl->server_local->crl_cache,
-                                 issuerhash, serialno))
+                                 issuerhash, serialno,
+                                 ctrl->force_crl_refresh))
         {
         case CRL_CACHE_VALID:
           err = 0;
@@ -426,19 +472,22 @@ cmd_isvalid (assuan_context_t ctx, char *line)
         case CRL_CACHE_DONTKNOW: 
           if (did_inquire)
             err = gpg_error (GPG_ERR_NO_CRL_KNOWN);
-          else if (!(err = inquire_cert (ctx)))
+          else if (!(err = inquire_cert_and_load_crl (ctx)))
             {
               did_inquire = 1;
               goto again;
             }
           break;
+        case CRL_CACHE_CANTUSE: 
+          err = gpg_error (GPG_ERR_NO_CRL_KNOWN);
+          break;
         default:
           log_fatal ("crl_cache_isvalid returned invalid code\n");
         }
     }
 
   if (err)
-    log_error ("command ISVALID failed: %s\n", gpg_strerror (err));
+    log_error (_("command %s failed: %s\n"), "ISVALID", gpg_strerror (err));
   xfree (issuerhash);
   return map_to_assuan_status (err);
 }
@@ -505,7 +554,7 @@ cmd_lookup (assuan_context_t ctx, char *line)
       err = start_cert_fetch (&fetch_context, list, ldapserver);
       if (err)
         {
-          log_error ("start_cert_fetch failed: %s\n", gpg_strerror (err));
+          log_error (_("start_cert_fetch failed: %s\n"), gpg_strerror (err));
           goto leave;
         }
 
@@ -531,7 +580,8 @@ cmd_lookup (assuan_context_t ctx, char *line)
             }
           if (err)
             {
-              log_error ("fetch_next_cert failed: %s\n", gpg_strerror (err));
+              log_error (_("fetch_next_cert failed: %s\n"),
+                         gpg_strerror (err));
               end_cert_fetch (fetch_context);
               goto leave;
             }
@@ -549,7 +599,7 @@ cmd_lookup (assuan_context_t ctx, char *line)
             ae = assuan_write_line (ctx, "END");
           if (ae) 
             {
-              log_error ("error sending data: %s\n", assuan_strerror (ae));
+              log_error (_("error sending data: %s\n"), assuan_strerror (ae));
               err = map_assuan_err (ae);
               end_cert_fetch (fetch_context);
               goto leave;
@@ -558,7 +608,7 @@ cmd_lookup (assuan_context_t ctx, char *line)
           if (++count >= opt.max_replies )
             {
               truncation_forced = 1;
-              log_info ("max_replies %d exceeded\n", opt.max_replies );
+              log_info (_("max_replies %d exceeded\n"), opt.max_replies );
             }
         }
 
@@ -575,7 +625,7 @@ cmd_lookup (assuan_context_t ctx, char *line)
 
  leave:
   if (err)
-    log_error ("Command LOOKUP failed: %s\n", gpg_strerror (err));
+    log_error (_("command %s failed: %s\n"), "LOOKUP", gpg_strerror (err));
   free_strlist (list);
   return map_to_assuan_status (err);
 }
@@ -603,10 +653,11 @@ cmd_loadcrl (assuan_context_t ctx, char *line)
   err = crl_cache_load (ctrl, ctrl->server_local->crl_cache, buf);
   xfree (buf);
   if (err)
-    log_error ("command LOADCRL failed: %s\n", gpg_strerror (err));
+    log_error (_("command %s failed: %s\n"), "LOADCRL", gpg_strerror (err));
   return map_to_assuan_status (err);
 }
 
+
 /* LISTCRLS 
 
    List the content of all CRLs in a readable format.  This command is
@@ -623,11 +674,11 @@ cmd_listcrls (assuan_context_t ctx, char *line)
   FILE *fp = assuan_get_data_fp (ctx);
 
   if (!fp)
-    return PARM_ERROR ("no data stream");
+    return PARM_ERROR (_("no data stream"));
 
   err = crl_cache_list (ctrl->server_local->crl_cache, fp);
   if (err)
-    log_error ("command LISTCRLS failed: %s\n", gpg_strerror (err));
+    log_error (_("command %s failed: %s\n"), "LISTCRLS", gpg_strerror (err));
   return map_to_assuan_status (err);
 }
 
@@ -694,7 +745,7 @@ start_command_handler (void)
       dirmngr_exit (2);
     }
   assuan_set_hello_line (ctx, "Dirmngr " VERSION " at your service");
-
+  assuan_register_option_handler (ctx, option_handler);
 
   assuan_set_pointer (ctx, &ctrl);
   ctrl.server_local = xcalloc (1, sizeof *ctrl.server_local);
@@ -705,6 +756,9 @@ start_command_handler (void)
   if (DBG_ASSUAN)
     assuan_set_log_stream (ctx, log_get_stream ());
 
+  if (!primary_connection)
+    primary_connection = &ctrl;
+
   for (;;) 
     {
       rc = assuan_accept (ctx);
@@ -724,7 +778,74 @@ start_command_handler (void)
         }
     }
   
+  if (primary_connection == &ctrl)
+    primary_connection = NULL;
+
   crl_cache_deinit (ctrl.server_local->crl_cache);
 
+
   assuan_deinit_server (ctx);
 }
+
+
+/* Send a status line back to the client.  KEYWORD is the status
+   keyword, the optioal string argumenst are blank separated added to
+   the line, the last argument must be a NULL. */
+void
+dirmngr_status (ctrl_t ctrl, const char *keyword, ...)
+{
+  va_list arg_ptr;
+  const char *text;
+
+  va_start (arg_ptr, keyword);
+
+  if (ctrl->server_local)
+    {
+      assuan_context_t ctx = ctrl->server_local->assuan_ctx;
+      char buf[950], *p;
+      size_t n;
+      
+      p = buf; 
+      n = 0;
+      while ( (text = va_arg (arg_ptr, const char *)) )
+        {
+          if (n)
+            {
+              *p++ = ' ';
+              n++;
+            }
+          for ( ; *text && n < DIM (buf)-2; n++)
+            *p++ = *text++;
+        }
+      *p = 0;
+      assuan_write_status (ctx, keyword, buf);
+    }
+
+  va_end (arg_ptr);
+}
+
+
+/* Note, that we ignore CTRL for now but use the first connection to
+   send the progress info back. */
+void
+dirmngr_tick (ctrl_t ctrl)
+{
+  static clock_t next_tick = 0;
+  clock_t now = clock ();
+
+  if (!next_tick)
+    {
+      next_tick = now + CLOCKS_PER_SEC;
+      if (next_tick < now)
+        next_tick = 0;
+    }
+  else if ( now > next_tick )
+    {
+      next_tick = now + CLOCKS_PER_SEC;
+      if (next_tick < now)
+        next_tick = 0;
+      
+      if (primary_connection)
+        dirmngr_status (primary_connection, "PROGRESS", "tick", "? 0 0", NULL);
+    }
+}
